@@ -1,6 +1,6 @@
 /*
     FUSE: Filesystem in Userspace
-    Copyright (C) 2001-2005  Miklos Szeredi <miklos@szeredi.hu>
+    Copyright (C) 2001-2006  Miklos Szeredi <miklos@szeredi.hu>
 
     This program can be distributed under the terms of the GNU LGPL.
     See the file COPYING.LIB.
@@ -27,6 +27,8 @@ struct fuse_worker {
     struct fuse_chan *ch;
     struct fuse_chan *prevch;
     pthread_t threads[FUSE_MAX_WORKERS];
+    pthread_t main_thread;
+    int exit;
     int error;
 };
 
@@ -58,7 +60,6 @@ static int start_thread(struct fuse_worker *w, pthread_t *thread_id);
 static void *do_work(void *data)
 {
     struct fuse_worker *w = (struct fuse_worker *) data;
-    int is_mainthread = (w->numworker == 1);
     size_t bufsize = fuse_chan_bufsize(w->prevch);
     char *buf = (char *) malloc(bufsize);
     if (!buf) {
@@ -70,7 +71,7 @@ static void *do_work(void *data)
 
     pthread_cleanup_push(free, buf);
     pthread_setcancelstate(PTHREAD_CANCEL_ENABLE, NULL);
-    pthread_setcanceltype(PTHREAD_CANCEL_ASYNCHRONOUS, NULL);
+    pthread_setcanceltype(PTHREAD_CANCEL_DEFERRED, NULL);
 
     while (!fuse_session_exited(w->se)) {
         int res = fuse_chan_receive(w->prevch, buf, bufsize);
@@ -83,6 +84,10 @@ static void *do_work(void *data)
         }
 
         pthread_mutex_lock(&w->lock);
+        if (w->exit) {
+            pthread_mutex_unlock(&w->lock);
+            break;
+        }
         w->numavail--;
         if (w->numavail == 0 && w->numworker < FUSE_MAX_WORKERS) {
             if (w->numworker < FUSE_MAX_WORKERS) {
@@ -102,22 +107,34 @@ static void *do_work(void *data)
     }
     pthread_cleanup_pop(1);
 
-    /* Wait for cancellation */
-    if (!is_mainthread)
+    if (pthread_self() != w->main_thread) {
+        pthread_kill(w->main_thread, SIGTERM);
         pause();
+    }
 
     return NULL;
 }
 
 static int start_thread(struct fuse_worker *w, pthread_t *thread_id)
 {
-    int res = pthread_create(thread_id, NULL, do_work, w);
+    sigset_t oldset;
+    sigset_t newset;
+    int res;
+
+    /* Disallow signal reception in worker threads */
+    sigemptyset(&newset);
+    sigaddset(&newset, SIGTERM);
+    sigaddset(&newset, SIGINT);
+    sigaddset(&newset, SIGHUP);
+    sigaddset(&newset, SIGQUIT);
+    pthread_sigmask(SIG_BLOCK, &newset, &oldset);
+    res = pthread_create(thread_id, NULL, do_work, w);
+    pthread_sigmask(SIG_SETMASK, &oldset, NULL);
     if (res != 0) {
         fprintf(stderr, "fuse: error creating thread: %s\n", strerror(res));
         return -1;
     }
 
-    pthread_detach(*thread_id);
     return 0;
 }
 
@@ -138,7 +155,7 @@ int fuse_session_loop_mt(struct fuse_session *se)
     memset(w, 0, sizeof(struct fuse_worker));
     w->se = se;
     w->prevch = fuse_session_next_chan(se, NULL);
-    w->ch = fuse_chan_new(&cop, -1, 0, w);
+    w->ch = fuse_chan_new(&cop, -1, fuse_chan_bufsize(w->prevch), w);
     if (w->ch == NULL) {
         free(w);
         return -1;
@@ -146,6 +163,7 @@ int fuse_session_loop_mt(struct fuse_session *se)
     w->error = 0;
     w->numworker = 1;
     w->numavail = 1;
+    w->main_thread = pthread_self();
     mutex_init(&w->lock);
 
     do_work(w);
@@ -153,7 +171,10 @@ int fuse_session_loop_mt(struct fuse_session *se)
     pthread_mutex_lock(&w->lock);
     for (i = 1; i < w->numworker; i++)
         pthread_cancel(w->threads[i]);
+    w->exit = 1;
     pthread_mutex_unlock(&w->lock);
+    for (i = 1; i < w->numworker; i++)
+        pthread_join(w->threads[i], NULL);
     pthread_mutex_destroy(&w->lock);
     err = w->error;
     fuse_chan_destroy(w->ch);
