@@ -1,6 +1,6 @@
 /*
     FUSE: Filesystem in Userspace
-    Copyright (C) 2001-2005  Miklos Szeredi <miklos@szeredi.hu>
+    Copyright (C) 2001-2006  Miklos Szeredi <miklos@szeredi.hu>
 
     This program can be distributed under the terms of the GNU LGPL.
     See the file COPYING.LIB
@@ -9,20 +9,21 @@
 #include <config.h>
 #include "fuse_lowlevel.h"
 #include "fuse_kernel.h"
+#include "fuse_opt.h"
 
 #include <stdio.h>
-#include <string.h>
 #include <stdlib.h>
+#include <stddef.h>
+#include <string.h>
 #include <unistd.h>
 #include <limits.h>
 #include <errno.h>
-#include <stdint.h>
 
 #define PARAM(inarg) (((char *)(inarg)) + sizeof(*(inarg)))
 
 struct fuse_ll {
-    unsigned int debug : 1;
-    unsigned int allow_root : 1;
+    int debug;
+    int allow_root;
     struct fuse_lowlevel_ops op;
     int got_init;
     void *userdata;
@@ -69,6 +70,8 @@ static const char *opname(enum fuse_opcode opcode)
     case FUSE_READDIR:		return "READDIR";
     case FUSE_RELEASEDIR:	return "RELEASEDIR";
     case FUSE_FSYNCDIR:		return "FSYNCDIR";
+    case FUSE_ACCESS:		return "ACCESS";
+    case FUSE_CREATE:		return "CREATE";
     default: 			return "???";
     }
 }
@@ -93,7 +96,7 @@ static void convert_stat(const struct stat *stbuf, struct fuse_attr *attr)
 #endif
 }
 
-static void convert_attr(const struct fuse_attr *attr, struct stat *stbuf)
+static void convert_attr(const struct fuse_setattr_in *attr, struct stat *stbuf)
 {
     stbuf->st_mode         = attr->mode;
     stbuf->st_uid          = attr->uid;
@@ -101,11 +104,9 @@ static void convert_attr(const struct fuse_attr *attr, struct stat *stbuf)
     stbuf->st_size         = attr->size;
     stbuf->st_atime        = attr->atime;
     stbuf->st_mtime        = attr->mtime;
-    stbuf->st_ctime        = attr->ctime;
 #ifdef HAVE_STRUCT_STAT_ST_ATIM
     stbuf->st_atim.tv_nsec = attr->atimensec;
     stbuf->st_mtim.tv_nsec = attr->mtimensec;
-    stbuf->st_ctim.tv_nsec = attr->ctimensec;
 #endif
 }
 
@@ -185,16 +186,17 @@ char *fuse_add_dirent(char *buf, const char *name, const struct stat *stbuf,
     return buf + entsize;
 }
 
-static void convert_statfs(const struct statfs *stbuf,
+static void convert_statfs(const struct statvfs *stbuf,
                            struct fuse_kstatfs *kstatfs)
 {
     kstatfs->bsize	= stbuf->f_bsize;
+    kstatfs->frsize	= stbuf->f_frsize;
     kstatfs->blocks	= stbuf->f_blocks;
     kstatfs->bfree	= stbuf->f_bfree;
     kstatfs->bavail	= stbuf->f_bavail;
     kstatfs->files	= stbuf->f_files;
     kstatfs->ffree	= stbuf->f_ffree;
-    kstatfs->namelen	= stbuf->f_namelen;
+    kstatfs->namelen	= stbuf->f_namemax;
 }
 
 static int send_reply_ok(fuse_req_t req, const void *arg, size_t argsize)
@@ -259,11 +261,30 @@ int fuse_reply_entry(fuse_req_t req, const struct fuse_entry_param *e)
 {
     struct fuse_entry_out arg;
 
+    /* before ABI 7.4 e->ino == 0 was invalid, only ENOENT meant
+       negative entry */
+    if (!e->ino && req->f->minor < 4)
+        return fuse_reply_err(req, ENOENT);
+
     memset(&arg, 0, sizeof(arg));
     fill_entry(&arg, e);
     return send_reply_ok(req, &arg, sizeof(arg));
 }
 
+int fuse_reply_create(fuse_req_t req, const struct fuse_entry_param *e,
+                      const struct fuse_file_info *f)
+{
+    struct {
+        struct fuse_entry_out e;
+        struct fuse_open_out o;
+    } arg;
+
+    memset(&arg, 0, sizeof(arg));
+    fill_entry(&arg.e, e);
+    fill_open(&arg.o, f);
+    return send_reply_ok(req, &arg, sizeof(arg));
+}
+
 int fuse_reply_attr(fuse_req_t req, const struct stat *attr,
                     double attr_timeout)
 {
@@ -306,14 +327,15 @@ int fuse_reply_buf(fuse_req_t req, const char *buf, size_t size)
     return send_reply_ok(req, buf, size);
 }
 
-int fuse_reply_statfs(fuse_req_t req, const struct statfs *stbuf)
+int fuse_reply_statfs(fuse_req_t req, const struct statvfs *stbuf)
 {
     struct fuse_statfs_out arg;
+    size_t size = req->f->minor < 4 ? FUSE_COMPAT_STATFS_SIZE : sizeof(arg);
 
     memset(&arg, 0, sizeof(arg));
     convert_statfs(stbuf, &arg.st);
 
-    return send_reply_ok(req, &arg, sizeof(arg));
+    return send_reply_ok(req, &arg, size);
 }
 
 int fuse_reply_xattr(fuse_req_t req, size_t count)
@@ -350,17 +372,35 @@ static void do_getattr(fuse_req_t req, fuse_ino_t nodeid)
 }
 
 static void do_setattr(fuse_req_t req, fuse_ino_t nodeid,
-                       struct fuse_setattr_in *arg, struct fuse_file_info *fi)
+                       struct fuse_setattr_in *arg)
 {
     if (req->f->op.setattr) {
+        struct fuse_file_info *fi = NULL;
+        struct fuse_file_info fi_store;
         struct stat stbuf;
         memset(&stbuf, 0, sizeof(stbuf));
-        convert_attr(&arg->attr, &stbuf);
+        convert_attr(arg, &stbuf);
+        if (arg->valid & FATTR_FH) {
+            arg->valid &= ~FATTR_FH;
+            memset(&fi_store, 0, sizeof(fi_store));
+            fi = &fi_store;
+            fi->fh = arg->fh;
+            fi->fh_old = fi->fh;
+        }
         req->f->op.setattr(req, nodeid, &stbuf, arg->valid, fi);
     } else
         fuse_reply_err(req, ENOSYS);
 }
 
+static void do_access(fuse_req_t req, fuse_ino_t nodeid,
+                      struct fuse_access_in *arg)
+{
+    if (req->f->op.access)
+        req->f->op.access(req, nodeid, arg->mask);
+    else
+        fuse_reply_err(req, ENOSYS);
+}
+
 static void do_readlink(fuse_req_t req, fuse_ino_t nodeid)
 {
     if (req->f->op.readlink)
@@ -433,6 +473,20 @@ static void do_link(fuse_req_t req, fuse_ino_t nodeid,
         fuse_reply_err(req, ENOSYS);
 }
 
+static void do_create(fuse_req_t req, fuse_ino_t nodeid,
+                      struct fuse_open_in *arg)
+{
+    if (req->f->op.create) {
+        struct fuse_file_info fi;
+
+        memset(&fi, 0, sizeof(fi));
+        fi.flags = arg->flags;
+
+        req->f->op.create(req, nodeid, PARAM(arg), arg->mode, &fi);
+    } else
+        fuse_reply_err(req, ENOSYS);
+}
+
 static void do_open(fuse_req_t req, fuse_ino_t nodeid,
                     struct fuse_open_in *arg)
 {
@@ -455,6 +509,7 @@ static void do_read(fuse_req_t req, fuse_ino_t nodeid,
 
         memset(&fi, 0, sizeof(fi));
         fi.fh = arg->fh;
+        fi.fh_old = fi.fh;
         req->f->op.read(req, nodeid, arg->size, arg->offset, &fi);
     } else
         fuse_reply_err(req, ENOSYS);
@@ -467,6 +522,7 @@ static void do_write(fuse_req_t req, fuse_ino_t nodeid,
 
     memset(&fi, 0, sizeof(fi));
     fi.fh = arg->fh;
+    fi.fh_old = fi.fh;
     fi.writepage = arg->write_flags & 1;
 
     if (req->f->op.write)
@@ -483,6 +539,7 @@ static void do_flush(fuse_req_t req, fuse_ino_t nodeid,
 
     memset(&fi, 0, sizeof(fi));
     fi.fh = arg->fh;
+    fi.fh_old = fi.fh;
 
     if (req->f->op.flush)
         req->f->op.flush(req, nodeid, &fi);
@@ -498,6 +555,7 @@ static void do_release(fuse_req_t req, fuse_ino_t nodeid,
     memset(&fi, 0, sizeof(fi));
     fi.flags = arg->flags;
     fi.fh = arg->fh;
+    fi.fh_old = fi.fh;
 
     if (req->f->op.release)
         req->f->op.release(req, nodeid, &fi);
@@ -512,6 +570,7 @@ static void do_fsync(fuse_req_t req, fuse_ino_t nodeid,
 
     memset(&fi, 0, sizeof(fi));
     fi.fh = inarg->fh;
+    fi.fh_old = fi.fh;
 
     if (req->f->op.fsync)
         req->f->op.fsync(req, nodeid, inarg->fsync_flags & 1, &fi);
@@ -540,6 +599,7 @@ static void do_readdir(fuse_req_t req, fuse_ino_t nodeid,
 
     memset(&fi, 0, sizeof(fi));
     fi.fh = arg->fh;
+    fi.fh_old = fi.fh;
 
     if (req->f->op.readdir)
         req->f->op.readdir(req, nodeid, arg->size, arg->offset, &fi);
@@ -555,6 +615,7 @@ static void do_releasedir(fuse_req_t req, fuse_ino_t nodeid,
     memset(&fi, 0, sizeof(fi));
     fi.flags = arg->flags;
     fi.fh = arg->fh;
+    fi.fh_old = fi.fh;
 
     if (req->f->op.releasedir)
         req->f->op.releasedir(req, nodeid, &fi);
@@ -569,6 +630,7 @@ static void do_fsyncdir(fuse_req_t req, fuse_ino_t nodeid,
 
     memset(&fi, 0, sizeof(fi));
     fi.fh = inarg->fh;
+    fi.fh_old = fi.fh;
 
     if (req->f->op.fsyncdir)
         req->f->op.fsyncdir(req, nodeid, inarg->fsync_flags & 1, &fi);
@@ -623,10 +685,11 @@ static void do_removexattr(fuse_req_t req, fuse_ino_t nodeid, char *name)
         fuse_reply_err(req, ENOSYS);
 }
 
-static void do_init(fuse_req_t req, struct fuse_init_in_out *arg)
+static void do_init(fuse_req_t req, struct fuse_init_in *arg)
 {
-    struct fuse_init_in_out outarg;
+    struct fuse_init_out outarg;
     struct fuse_ll *f = req->f;
+    size_t bufsize = fuse_chan_bufsize(req->ch);
 
     if (f->debug) {
         printf("INIT: %u.%u\n", arg->major, arg->minor);
@@ -637,18 +700,24 @@ static void do_init(fuse_req_t req, struct fuse_init_in_out *arg)
         f->op.init(f->userdata);
 
     f->major = FUSE_KERNEL_VERSION;
-    f->minor = FUSE_KERNEL_MINOR_VERSION;
+    f->minor = arg->minor;
+
+    if (bufsize < FUSE_MIN_READ_BUFFER) {
+        fprintf(stderr, "fuse: warning: buffer size too small: %i\n", bufsize);
+        bufsize = FUSE_MIN_READ_BUFFER;
+    }
 
     memset(&outarg, 0, sizeof(outarg));
     outarg.major = f->major;
-    outarg.minor = f->minor;
+    outarg.minor = FUSE_KERNEL_MINOR_VERSION;
+    outarg.max_write = bufsize - 4096;
 
     if (f->debug) {
         printf("   INIT: %u.%u\n", outarg.major, outarg.minor);
         fflush(stdout);
     }
 
-    send_reply_ok(req, &outarg, sizeof(outarg));
+    send_reply_ok(req, &outarg, arg->minor < 5 ? 8 : sizeof(outarg));
 }
 
 void *fuse_req_userdata(fuse_req_t req)
@@ -690,7 +759,7 @@ static void fuse_ll_process(void *data, const char *buf, size_t len,
     req->ch = ch;
 
     if (!f->got_init && in->opcode != FUSE_INIT)
-        fuse_reply_err(req, EPROTO);
+        fuse_reply_err(req, EIO);
     else if (f->allow_root && in->uid != f->owner && in->uid != 0 &&
              in->opcode != FUSE_INIT && in->opcode != FUSE_READ &&
              in->opcode != FUSE_WRITE && in->opcode != FUSE_FSYNC &&
@@ -699,7 +768,7 @@ static void fuse_ll_process(void *data, const char *buf, size_t len,
         fuse_reply_err(req, EACCES);
     } else switch (in->opcode) {
     case FUSE_INIT:
-        do_init(req, (struct fuse_init_in_out *) inarg);
+        do_init(req, (struct fuse_init_in *) inarg);
         break;
 
     case FUSE_LOOKUP:
@@ -715,7 +784,7 @@ static void fuse_ll_process(void *data, const char *buf, size_t len,
         break;
 
     case FUSE_SETATTR:
-        do_setattr(req, in->nodeid, (struct fuse_setattr_in *) inarg, NULL);
+        do_setattr(req, in->nodeid, (struct fuse_setattr_in *) inarg);
         break;
 
     case FUSE_READLINK:
@@ -811,43 +880,64 @@ static void fuse_ll_process(void *data, const char *buf, size_t len,
         do_fsyncdir(req, in->nodeid, (struct fuse_fsync_in *) inarg);
         break;
 
+    case FUSE_ACCESS:
+        do_access(req, in->nodeid, (struct fuse_access_in *) inarg);
+        break;
+
+    case FUSE_CREATE:
+        do_create(req, in->nodeid, (struct fuse_open_in *) inarg);
+        break;
+
     default:
         fuse_reply_err(req, ENOSYS);
     }
 }
 
-int fuse_lowlevel_is_lib_option(const char *opt)
+enum {
+    KEY_HELP,
+    KEY_VERSION,
+};
+
+static struct fuse_opt fuse_ll_opts[] = {
+    { "debug", offsetof(struct fuse_ll, debug), 1 },
+    { "-d", offsetof(struct fuse_ll, debug), 1 },
+    { "allow_root", offsetof(struct fuse_ll, allow_root), 1 },
+    FUSE_OPT_KEY("-h", KEY_HELP),
+    FUSE_OPT_KEY("--help", KEY_HELP),
+    FUSE_OPT_KEY("-V", KEY_VERSION),
+    FUSE_OPT_KEY("--version", KEY_VERSION),
+    FUSE_OPT_END
+};
+
+static void fuse_ll_version(void)
 {
-    if (strcmp(opt, "debug") == 0 ||
-        strcmp(opt, "allow_root") == 0)
-        return 1;
-    else
-        return 0;
+    fprintf(stderr, "using FUSE kernel interface version %i.%i\n",
+            FUSE_KERNEL_VERSION, FUSE_KERNEL_MINOR_VERSION);
 }
 
-static int parse_ll_opts(struct fuse_ll *f, const char *opts)
+static int fuse_ll_opt_proc(void *data, const char *arg, int key,
+                            struct fuse_args *outargs)
 {
-    if (opts) {
-        char *xopts = strdup(opts);
-        char *s = xopts;
-        char *opt;
+    (void) data; (void) outargs;
 
-        if (xopts == NULL) {
-            fprintf(stderr, "fuse: memory allocation failed\n");
-            return -1;
-        }
+    switch (key) {
+    case KEY_HELP:
+        break;
 
-        while((opt = strsep(&s, ","))) {
-            if (strcmp(opt, "debug") == 0)
-                f->debug = 1;
-            else if (strcmp(opt, "allow_root") == 0)
-                f->allow_root = 1;
-            else
-                fprintf(stderr, "fuse: warning: unknown option `%s'\n", opt);
-        }
-        free(xopts);
+    case KEY_VERSION:
+        fuse_ll_version();
+        break;
+
+    default:
+        fprintf(stderr, "fuse: unknown option `%s'\n", arg);
     }
-    return 0;
+
+    return -1;
+}
+
+int fuse_lowlevel_is_lib_option(const char *opt)
+{
+    return fuse_opt_match(fuse_ll_opts, opt);
 }
 
 static void fuse_ll_destroy(void *data)
@@ -860,8 +950,7 @@ static void fuse_ll_destroy(void *data)
     free(f);
 }
 
-
-struct fuse_session *fuse_lowlevel_new(const char *opts,
+struct fuse_session *fuse_lowlevel_new(struct fuse_args *args,
                                        const struct fuse_lowlevel_ops *op,
                                        size_t op_size, void *userdata)
 {
@@ -883,7 +972,7 @@ struct fuse_session *fuse_lowlevel_new(const char *opts,
         goto out;
     }
 
-    if (parse_ll_opts(f, opts) == -1)
+    if (fuse_opt_parse(args, f, fuse_ll_opts, fuse_ll_opt_proc) == -1)
         goto out_free;
 
     memcpy(&f->op, op, op_size);
@@ -901,3 +990,75 @@ struct fuse_session *fuse_lowlevel_new(const char *opts,
  out:
     return NULL;
 }
+
+#ifndef __FreeBSD__
+
+#include "fuse_lowlevel_compat.h"
+
+static void fill_open_compat(struct fuse_open_out *arg,
+                      const struct fuse_file_info_compat *f)
+{
+    arg->fh = f->fh;
+    if (f->direct_io)
+        arg->open_flags |= FOPEN_DIRECT_IO;
+    if (f->keep_cache)
+        arg->open_flags |= FOPEN_KEEP_CACHE;
+}
+
+static void convert_statfs_compat(const struct statfs *compatbuf,
+                                  struct statvfs *buf)
+{
+    buf->f_bsize	= compatbuf->f_bsize;
+    buf->f_blocks	= compatbuf->f_blocks;
+    buf->f_bfree	= compatbuf->f_bfree;
+    buf->f_bavail	= compatbuf->f_bavail;
+    buf->f_files	= compatbuf->f_files;
+    buf->f_ffree	= compatbuf->f_ffree;
+    buf->f_namemax	= compatbuf->f_namelen;
+}
+
+int fuse_reply_open_compat(fuse_req_t req,
+                           const struct fuse_file_info_compat *f)
+{
+    struct fuse_open_out arg;
+
+    memset(&arg, 0, sizeof(arg));
+    fill_open_compat(&arg, f);
+    return send_reply_ok(req, &arg, sizeof(arg));
+}
+
+int fuse_reply_statfs_compat(fuse_req_t req, const struct statfs *stbuf)
+{
+    struct statvfs newbuf;
+
+    memset(&newbuf, 0, sizeof(newbuf));
+    convert_statfs_compat(stbuf, &newbuf);
+
+    return fuse_reply_statfs(req, &newbuf);
+}
+
+struct fuse_session *fuse_lowlevel_new_compat(const char *opts,
+                                              const struct fuse_lowlevel_ops *op,
+                                              size_t op_size, void *userdata)
+{
+    struct fuse_session *se;
+    struct fuse_args args = FUSE_ARGS_INIT(0, NULL);
+
+    if (opts &&
+        (fuse_opt_add_arg(&args, "") == -1 ||
+         fuse_opt_add_arg(&args, "-o") == -1 ||
+         fuse_opt_add_arg(&args, opts) == -1)) {
+        fuse_opt_free_args(&args);
+        return NULL;
+    }
+    se = fuse_lowlevel_new(&args, op, op_size, userdata);
+    fuse_opt_free_args(&args);
+
+    return se;
+}
+
+__asm__(".symver fuse_reply_statfs_compat,fuse_reply_statfs@FUSE_2.4");
+__asm__(".symver fuse_reply_open_compat,fuse_reply_open@FUSE_2.4");
+__asm__(".symver fuse_lowlevel_new_compat,fuse_lowlevel_new@FUSE_2.4");
+
+#endif /* __FreeBSD__ */
