@@ -1,6 +1,6 @@
 /*
     FUSE: Filesystem in Userspace
-    Copyright (C) 2001-2005  Miklos Szeredi <miklos@szeredi.hu>
+    Copyright (C) 2001-2006  Miklos Szeredi <miklos@szeredi.hu>
 
     This program can be distributed under the terms of the GNU GPL.
     See the file COPYING.
@@ -46,6 +46,10 @@
 #define FUSE_VERSION_FILE_OLD "/proc/fs/fuse/version"
 #define FUSE_CONF "/etc/fuse.conf"
 
+#ifndef MS_DIRSYNC
+#define MS_DIRSYNC 128
+#endif
+
 static const char *progname;
 
 static int user_allow_other = 0;
@@ -145,7 +149,7 @@ static int add_mount(const char *fsname, const char *mnt, const char *type,
     struct mntent ent;
     FILE *fp;
 
-    if (check_name(fsname) == -1 || check_name(mnt) == -1 || 
+    if (check_name(fsname) == -1 || check_name(mnt) == -1 ||
         check_name(type) == -1 || check_name(opts) == -1)
         return -1;
 
@@ -173,17 +177,40 @@ static int add_mount(const char *fsname, const char *mnt, const char *type,
     return 0;
 }
 
-static int remove_mount(const char *mnt, int quiet, const char *mtab,
-                        const char *mtab_new)
+static int unmount_rename(const char *mtab, const char *mtab_new)
+{
+    int res;
+    struct stat sbuf;
+
+    if (stat(mtab, &sbuf) == 0)
+        chown(mtab_new, sbuf.st_uid, sbuf.st_gid);
+
+    res = rename(mtab_new, mtab);
+    if (res == -1) {
+        fprintf(stderr, "%s: failed to rename %s to %s: %s\n", progname,
+                mtab_new, mtab, strerror(errno));
+        return -1;
+    }
+    return 0;
+}
+
+static int unmount_fuse(const char *mnt, int quiet, int lazy)
 {
     int res;
     struct mntent *entp;
     FILE *fp;
-    FILE *newfp;
+    FILE *newfp = NULL;
     const char *user = NULL;
     char uidstr[32];
     unsigned uidlen = 0;
     int found;
+    int issymlink = 0;
+    struct stat stbuf;
+    const char *mtab = _PATH_MOUNTED;
+    const char *mtab_new = _PATH_MOUNTED "~fuse~";
+
+    if (lstat(mtab, &stbuf) != -1 && S_ISLNK(stbuf.st_mode))
+        issymlink = 1;
 
     fp = setmntent(mtab, "r");
     if (fp == NULL) {
@@ -192,17 +219,20 @@ static int remove_mount(const char *mnt, int quiet, const char *mtab,
 	return -1;
     }
 
-    newfp = setmntent(mtab_new, "w");
-    if (newfp == NULL) {
-	fprintf(stderr, "%s: failed to open %s: %s\n", progname, mtab_new,
-		strerror(errno));
-	return -1;
+    if (!issymlink) {
+        newfp = setmntent(mtab_new, "w");
+        if (newfp == NULL) {
+            fprintf(stderr, "%s: failed to open %s: %s\n", progname, mtab_new,
+                    strerror(errno));
+            endmntent(fp);
+            return -1;
+        }
     }
 
     if (getuid() != 0) {
         user = get_user_name();
         if (user == NULL)
-            return -1;
+            goto err_endmntent;
 
         uidlen = sprintf(uidstr, "%u", getuid());
     }
@@ -229,7 +259,7 @@ static int remove_mount(const char *mnt, int quiet, const char *mtab,
         }
         if (removed)
             found = 1;
-        else {
+        else if (!issymlink) {
             res = addmntent(newfp, entp);
             if (res != 0) {
                 fprintf(stderr, "%s: failed to add entry to %s: %s\n",
@@ -239,17 +269,37 @@ static int remove_mount(const char *mnt, int quiet, const char *mtab,
     }
 
     endmntent(fp);
-    endmntent(newfp);
+    if (!issymlink)
+        endmntent(newfp);
 
     if (!found) {
         if (!quiet)
             fprintf(stderr, "%s: entry for %s not found in %s\n", progname,
                     mnt, mtab);
-        unlink(mtab_new);
-        return -1;
+        goto err;
     }
 
+    drop_privs();
+    res = do_unmount(mnt, quiet, lazy);
+    restore_privs();
+    if (res == -1)
+        goto err;
+
+    if (!issymlink) {
+        res = unmount_rename(mtab, mtab_new);
+        if (res == -1)
+            goto err;
+    }
     return 0;
+
+ err_endmntent:
+    if (!issymlink)
+        endmntent(newfp);
+    endmntent(fp);
+ err:
+    if (!issymlink)
+        unlink(mtab_new);
+    return -1;
 }
 
 static int count_fuse_fs(void)
@@ -271,47 +321,7 @@ static int count_fuse_fs(void)
     return count;
 }
 
-static int unmount_rename(const char *mnt, int quiet, int lazy,
-                          const char *mtab, const char *mtab_new)
-{
-    int res;
-    struct stat sbuf;
 
-    drop_privs();
-    res = do_unmount(mnt, quiet, lazy);
-    restore_privs();
-    if (res == -1)
-        return -1;
-
-    if (stat(mtab, &sbuf) == 0)
-        chown(mtab_new, sbuf.st_uid, sbuf.st_gid);
-
-    res = rename(mtab_new, mtab);
-    if (res == -1) {
-        fprintf(stderr, "%s: failed to rename %s to %s: %s\n", progname,
-                mtab_new, mtab, strerror(errno));
-        return -1;
-    }
-    return 0;
-}
-
-static int unmount_fuse(const char *mnt, int quiet, int lazy)
-{
-    int res;
-    const char *mtab = _PATH_MOUNTED;
-    const char *mtab_new = _PATH_MOUNTED "~fuse~";
-
-    res = remove_mount(mnt, quiet, mtab, mtab_new);
-    if (res == -1)
-        return -1;
-
-    res = unmount_rename(mnt, quiet, lazy, mtab, mtab_new);
-    if (res == -1) {
-        unlink(mtab_new);
-        return -1;
-    }
-    return 0;
-}
 #else /* IGNORE_MTAB */
 static int lock_mtab()
 {
@@ -425,6 +435,7 @@ static struct mount_flags mount_flags[] = {
     {"sync",    MS_SYNCHRONOUS, 1, 1},
     {"atime",   MS_NOATIME,     0, 1},
     {"noatime", MS_NOATIME,     1, 1},
+    {"dirsync", MS_DIRSYNC,     1, 1},
     {NULL,      0,              0, 0}
 };
 
@@ -1012,7 +1023,7 @@ static void usage(void)
             "%s: [options] mountpoint\n"
             "Options:\n"
             " -h                print help\n"
-            " -v                print version\n"
+            " -V                print version\n"
             " -o opt[,opt...]   mount options\n"
             " -u                unmount\n"
             " -q                quiet\n"
@@ -1023,7 +1034,7 @@ static void usage(void)
 
 static void show_version(void)
 {
-    printf("%s\n", PACKAGE_STRING);
+    printf("fusermount version: %s\n", PACKAGE_VERSION);
     exit(0);
 }
 
@@ -1046,7 +1057,7 @@ int main(int argc, char *argv[])
         {"lazy",    no_argument, NULL, 'z'},
         {"quiet",   no_argument, NULL, 'q'},
         {"help",    no_argument, NULL, 'h'},
-        {"version", no_argument, NULL, 'v'},
+        {"version", no_argument, NULL, 'V'},
         {0, 0, 0, 0}};
 
     progname = strdup(argv[0]);
@@ -1055,13 +1066,13 @@ int main(int argc, char *argv[])
         exit(1);
     }
 
-    while ((ch = getopt_long(argc, argv, "hvo:uzq", long_opts, NULL)) != -1) {
+    while ((ch = getopt_long(argc, argv, "hVo:uzq", long_opts, NULL)) != -1) {
         switch (ch) {
         case 'h':
             usage();
             break;
 
-        case 'v':
+        case 'V':
             show_version();
             break;
 
